@@ -6,28 +6,47 @@ from collections.abc import Sequence
 from collections.abc import Sized
 from functools import wraps
 from http import client
-from http import cookies
 from itertools import groupby
 from re import Pattern
+from threading import Lock as _ThreadingLock
+from typing import TYPE_CHECKING
+from typing import Any
+from typing import Callable
+from typing import Dict
+from typing import Iterable
+from typing import Iterator
+from typing import List
+from typing import Mapping
+from typing import Optional
+from typing import Tuple
+from typing import Type
+from typing import Union
+from typing import overload
 from warnings import warn
 
 from requests.adapters import HTTPAdapter
+from requests.adapters import MaxRetryError
 from requests.exceptions import ConnectionError
-from requests.utils import cookiejar_from_dict
 
 from responses.matchers import json_params_matcher as _json_params_matcher
 from responses.matchers import query_string_matcher as _query_string_matcher
 from responses.matchers import urlencoded_params_matcher as _urlencoded_params_matcher
 from responses.registries import FirstMatchRegistry
 
+try:
+    from typing_extensions import Literal
+except ImportError:  # pragma: no cover
+    from typing import Literal  # type: ignore  # pragma: no cover
+
 try:
     from requests.packages.urllib3.response import HTTPResponse
 except ImportError:  # pragma: no cover
     from urllib3.response import HTTPResponse  # pragma: no cover
+
 try:
     from requests.packages.urllib3.connection import HTTPHeaderDict
 except ImportError:  # pragma: no cover
-    from urllib3.response import HTTPHeaderDict  # pragma: no cover
+    from urllib3.response import HTTPHeaderDict  # type: ignore[attr-defined]
 try:
     from requests.packages.urllib3.util.url import parse_url
 except ImportError:  # pragma: no cover
@@ -42,6 +61,21 @@ from urllib.parse import urlsplit
 from urllib.parse import urlunparse
 from urllib.parse import urlunsplit
 
+if TYPE_CHECKING:  # pragma: no cover
+    # import only for linter run
+    from mypy.typeshed.stdlib.unittest.mock import _patcher as _mock_patcher
+    from requests import PreparedRequest
+    from requests import models
+    from urllib3 import Retry as _Retry
+
+# Block of type annotations
+_Body = Union[str, BaseException, "Response", BufferedReader, bytes, None]
+_F = Callable[..., Any]
+_HeaderSet = Optional[Union[Mapping[str, str], List[Tuple[str, str]]]]
+_MatcherIterable = Iterable[Callable[..., Tuple[bool, str]]]
+_HTTPMethodOrResponse = Optional[Union[str, "BaseResponse"]]
+_URLPatternType = Union["Pattern[str]", str]
+
 Call = namedtuple("Call", ["request", "response"])
 _real_send = HTTPAdapter.send
 _UNSET = object()
@@ -56,13 +90,13 @@ class FalseBool:
     https://github.com/getsentry/responses/issues/464
     """
 
-    def __bool__(self):
+    def __bool__(self) -> bool:
         return False
 
     __nonzero__ = __bool__
 
 
-def urlencoded_params_matcher(params):
+def urlencoded_params_matcher(params: Optional[Dict[str, str]]) -> Callable[..., Any]:
     warn(
         "Function is deprecated. Use 'from responses.matchers import urlencoded_params_matcher'",
         DeprecationWarning,
@@ -70,7 +104,7 @@ def urlencoded_params_matcher(params):
     return _urlencoded_params_matcher(params)
 
 
-def json_params_matcher(params):
+def json_params_matcher(params: Optional[Dict[str, Any]]) -> Callable[..., Any]:
     warn(
         "Function is deprecated. Use 'from responses.matchers import json_params_matcher'",
         DeprecationWarning,
@@ -78,11 +112,11 @@ def json_params_matcher(params):
     return _json_params_matcher(params)
 
 
-def _has_unicode(s):
+def _has_unicode(s: str) -> bool:
     return any(ord(char) > 128 for char in s)
 
 
-def _clean_unicode(url):
+def _clean_unicode(url: str) -> str:
     """Clean up URLs, which use punycode to handle unicode chars.
 
     Applies percent encoding to URL path and query if required.
@@ -118,30 +152,13 @@ def _clean_unicode(url):
     return "".join(chars)
 
 
-def _cookies_from_headers(headers):
-    """Create Cookies from request Headers.
-
-    Converts ``set-cookie`` headers to real cookies.
-
-    Parameters
-    ----------
-    headers : dict
-        Request headers.
-
-    Returns
-    -------
-    CookieJar
-        CookieJar request object.
-
-    """
-    resp_cookie = cookies.SimpleCookie()
-    resp_cookie.load(headers["set-cookie"])
-    cookies_dict = {name: v.value for name, v in resp_cookie.items()}
-
-    return cookiejar_from_dict(cookies_dict)
-
-
-def get_wrapped(func, responses, registry=None):
+def get_wrapped(
+    func: Callable[..., Any],
+    responses: "RequestsMock",
+    *,
+    registry: Optional[Any] = None,
+    assert_all_requests_are_fired: Optional[bool] = None,
+) -> Callable[..., Any]:
     """Wrap provided function inside ``responses`` context manager.
 
     Provides a synchronous or asynchronous wrapper for the function.
@@ -155,6 +172,8 @@ def get_wrapped(func, responses, registry=None):
         Mock object that is used as context manager.
     registry : FirstMatchRegistry, optional
         Custom registry that should be applied. See ``responses.registries``
+    assert_all_requests_are_fired : bool
+        Raise an error if not all registered responses were executed.
 
     Returns
     -------
@@ -165,53 +184,105 @@ def get_wrapped(func, responses, registry=None):
     if registry is not None:
         responses._set_registry(registry)
 
+    assert_mock = std_mock.patch.object(
+        target=responses,
+        attribute="assert_all_requests_are_fired",
+        new=assert_all_requests_are_fired,
+    )
+
     if inspect.iscoroutinefunction(func):
         # set asynchronous wrapper if requestor function is asynchronous
         @wraps(func)
-        async def wrapper(*args, **kwargs):
-            with responses:
+        async def wrapper(*args: Any, **kwargs: Any) -> Any:  # type: ignore[misc]
+
+            with assert_mock, responses:
                 return await func(*args, **kwargs)
 
     else:
 
         @wraps(func)
-        def wrapper(*args, **kwargs):
-            with responses:
+        def wrapper(*args: Any, **kwargs: Any) -> Any:  # type: ignore[misc]
+
+            with assert_mock, responses:
+                # set 'assert_all_requests_are_fired' temporarily for a single run.
+                # Mock automatically unsets to avoid leakage to another decorated
+                # function since we still apply the value on 'responses.mock' object
                 return func(*args, **kwargs)
 
     return wrapper
 
 
-class CallList(Sequence, Sized):
-    def __init__(self):
-        self._calls = []
+class CallList(Sequence, Sized):  # type: ignore[type-arg]
+    def __init__(self) -> None:
+        self._calls: List[Call] = []
 
-    def __iter__(self):
+    def __iter__(self) -> Iterator[Call]:
         return iter(self._calls)
 
-    def __len__(self):
+    def __len__(self) -> int:
         return len(self._calls)
 
-    def __getitem__(self, idx):
+    def __getitem__(self, idx: int) -> Call:  # type: ignore[override]
         return self._calls[idx]
 
-    def add(self, request, response):
+    def add(self, request: "PreparedRequest", response: _Body) -> None:
         self._calls.append(Call(request, response))
 
-    def reset(self):
+    def reset(self) -> None:
         self._calls = []
 
 
-def _ensure_url_default_path(url):
+def _ensure_url_default_path(
+    url: _URLPatternType,
+) -> _URLPatternType:
+    """Add empty URL path '/' if doesn't exist.
+
+    Examples
+    --------
+    >>> _ensure_url_default_path("http://example.com")
+    "http://example.com/"
+
+    Parameters
+    ----------
+    url : str or re.Pattern
+        URL to validate.
+
+    Returns
+    -------
+    url : str or re.Pattern
+        Modified URL if str or unchanged re.Pattern
+
+    """
     if isinstance(url, str):
         url_parts = list(urlsplit(url))
         if url_parts[2] == "":
             url_parts[2] = "/"
-        url = urlunsplit(url_parts)
+            url = urlunsplit(url_parts)
     return url
 
 
-def _get_url_and_path(url):
+def _get_url_and_path(url: str) -> str:
+    """Construct URL only containing scheme, netloc and path by truncating other parts.
+
+    This method complies with RFC 3986.
+
+    Examples
+    --------
+    >>> _get_url_and_path("http://example.com/path;segment?ab=xy&zed=qwe#test=1&foo=bar")
+    "http://example.com/path;segment"
+
+
+    Parameters
+    ----------
+    url : str
+        URL to parse.
+
+    Returns
+    -------
+    url : str
+        URL with scheme, netloc and path
+
+    """
     url_parsed = urlsplit(url)
     url_and_path = urlunparse(
         [url_parsed.scheme, url_parsed.netloc, url_parsed.path, None, None, None]
@@ -219,15 +290,30 @@ def _get_url_and_path(url):
     return parse_url(url_and_path).url
 
 
-def _handle_body(body):
+def _handle_body(
+    body: Optional[Union[bytes, BufferedReader, str]]
+) -> Union[BufferedReader, BytesIO]:
+    """Generates `Response` body.
+
+    Parameters
+    ----------
+    body : str or bytes or BufferedReader
+        Input data to generate `Response` body.
+
+    Returns
+    -------
+    body : BufferedReader or BytesIO
+        `Response` body
+
+    """
     if isinstance(body, str):
         body = body.encode("utf-8")
     if isinstance(body, BufferedReader):
         return body
 
-    data = BytesIO(body)
+    data = BytesIO(body)  # type: ignore[arg-type]
 
-    def is_closed():
+    def is_closed() -> bool:
         """
         Real Response uses HTTPResponse as body object.
         Thus, when method is_closed is called first to check if there is any more
@@ -252,28 +338,36 @@ def _handle_body(body):
             # only if file really closed (by us) return True
             return True
 
-    data.isclosed = is_closed
+    data.isclosed = is_closed  # type: ignore[attr-defined]
     return data
 
 
 class BaseResponse(object):
-    passthrough = False
-    content_type = None
-    headers = None
-    stream = False
+    passthrough: bool = False
+    content_type: Optional[str] = None
+    headers: Optional[Mapping[str, str]] = None
+    stream: Optional[bool] = False
 
-    def __init__(self, method, url, match_querystring=None, match=()):
-        self.method = method
+    def __init__(
+        self,
+        method: str,
+        url: _URLPatternType,
+        match_querystring: Union[bool, object] = None,
+        match: "_MatcherIterable" = (),
+    ) -> None:
+        self.method: str = method
         # ensure the url has a default path set if the url is a string
-        self.url = _ensure_url_default_path(url)
+        self.url: _URLPatternType = _ensure_url_default_path(url)
 
         if self._should_match_querystring(match_querystring):
-            match = tuple(match) + (_query_string_matcher(urlsplit(self.url).query),)
+            match = tuple(match) + (
+                _query_string_matcher(urlsplit(self.url).query),  # type: ignore[arg-type]
+            )
 
-        self.match = match
-        self.call_count = 0
+        self.match: "_MatcherIterable" = match
+        self.call_count: int = 0
 
-    def __eq__(self, other):
+    def __eq__(self, other: Any) -> bool:
         if not isinstance(other, BaseResponse):
             return False
 
@@ -288,10 +382,12 @@ class BaseResponse(object):
 
         return self_url == other_url
 
-    def __ne__(self, other):
+    def __ne__(self, other: Any) -> bool:
         return not self.__eq__(other)
 
-    def _should_match_querystring(self, match_querystring_argument):
+    def _should_match_querystring(
+        self, match_querystring_argument: Union[bool, object]
+    ) -> Union[bool, object]:
         if isinstance(self.url, Pattern):
             # the old default from <= 0.9.0
             return False
@@ -310,7 +406,7 @@ class BaseResponse(object):
 
         return bool(urlsplit(self.url).query)
 
-    def _url_matches(self, url, other):
+    def _url_matches(self, url: _URLPatternType, other: str) -> bool:
         if isinstance(url, str):
             if _has_unicode(url):
                 url = _clean_unicode(url)
@@ -324,7 +420,9 @@ class BaseResponse(object):
             return False
 
     @staticmethod
-    def _req_attr_matches(match, request):
+    def _req_attr_matches(
+        match: "_MatcherIterable", request: "PreparedRequest"
+    ) -> Tuple[bool, str]:
         for matcher in match:
             valid, reason = matcher(request)
             if not valid:
@@ -332,7 +430,7 @@ class BaseResponse(object):
 
         return True, ""
 
-    def get_headers(self):
+    def get_headers(self) -> HTTPHeaderDict:
         headers = HTTPHeaderDict()  # Duplicate headers are legal
         if self.content_type is not None:
             headers["Content-Type"] = self.content_type
@@ -340,14 +438,14 @@ class BaseResponse(object):
             headers.extend(self.headers)
         return headers
 
-    def get_response(self, request):
+    def get_response(self, request: "PreparedRequest") -> HTTPResponse:
         raise NotImplementedError
 
-    def matches(self, request):
+    def matches(self, request: "PreparedRequest") -> Tuple[bool, str]:
         if request.method != self.method:
             return False, "Method does not match"
 
-        if not self._url_matches(self.url, request.url):
+        if not self._url_matches(self.url, str(request.url)):
             return False, "URL does not match"
 
         valid, reason = self._req_attr_matches(self.match, request)
@@ -360,17 +458,17 @@ class BaseResponse(object):
 class Response(BaseResponse):
     def __init__(
         self,
-        method,
-        url,
-        body="",
-        json=None,
-        status=200,
-        headers=None,
-        stream=None,
-        content_type=_UNSET,
-        auto_calculate_content_length=False,
-        **kwargs,
-    ):
+        method: str,
+        url: _URLPatternType,
+        body: _Body = "",
+        json: Optional[Any] = None,
+        status: int = 200,
+        headers: Optional[Mapping[str, str]] = None,
+        stream: Optional[bool] = None,
+        content_type: Union[str, object] = _UNSET,
+        auto_calculate_content_length: bool = False,
+        **kwargs: Any,
+    ) -> None:
         # if we were passed a `json` argument,
         # override the body and content_type
         if json is not None:
@@ -385,9 +483,9 @@ class Response(BaseResponse):
             else:
                 content_type = "text/plain"
 
-        self.body = body
-        self.status = status
-        self.headers = headers
+        self.body: _Body = body
+        self.status: int = status
+        self.headers: Optional[Mapping[str, str]] = headers
 
         if stream is not None:
             warn(
@@ -395,17 +493,19 @@ class Response(BaseResponse):
                 DeprecationWarning,
             )
 
-        self.stream = stream
-        self.content_type = content_type
-        self.auto_calculate_content_length = auto_calculate_content_length
+        self.stream: Optional[bool] = stream
+        self.content_type: str = content_type  # type: ignore[assignment]
+        self.auto_calculate_content_length: bool = auto_calculate_content_length
         super().__init__(method, url, **kwargs)
 
-    def get_response(self, request):
+    def get_response(self, request: "PreparedRequest") -> HTTPResponse:
         if self.body and isinstance(self.body, Exception):
             raise self.body
 
         headers = self.get_headers()
         status = self.status
+
+        assert not isinstance(self.body, (Response, BaseException))
         body = _handle_body(self.body)
 
         if (
@@ -425,7 +525,7 @@ class Response(BaseResponse):
             preload_content=False,
         )
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return (
             "<Response(url='{url}' status={status} "
             "content_type='{content_type}' headers='{headers}')>".format(
@@ -439,8 +539,14 @@ class Response(BaseResponse):
 
 class CallbackResponse(BaseResponse):
     def __init__(
-        self, method, url, callback, stream=None, content_type="text/plain", **kwargs
-    ):
+        self,
+        method: str,
+        url: _URLPatternType,
+        callback: Callable[[Any], Any],
+        stream: Optional[bool] = None,
+        content_type: Optional[str] = "text/plain",
+        **kwargs: Any,
+    ) -> None:
         self.callback = callback
 
         if stream is not None:
@@ -448,11 +554,11 @@ class CallbackResponse(BaseResponse):
                 "stream argument is deprecated. Use stream parameter in request directly",
                 DeprecationWarning,
             )
-        self.stream = stream
-        self.content_type = content_type
+        self.stream: Optional[bool] = stream
+        self.content_type: Optional[str] = content_type
         super().__init__(method, url, **kwargs)
 
-    def get_response(self, request):
+    def get_response(self, request: "PreparedRequest") -> HTTPResponse:
         headers = self.get_headers()
 
         result = self.callback(request)
@@ -490,7 +596,7 @@ class CallbackResponse(BaseResponse):
 
 
 class PassthroughResponse(BaseResponse):
-    passthrough = True
+    passthrough: bool = True
 
 
 class OriginalResponseShim(object):
@@ -506,47 +612,49 @@ class OriginalResponseShim(object):
     [1]: https://github.com/psf/requests/blob/75bdc998e2d/requests/cookies.py#L125
     """
 
-    def __init__(self, headers):
-        self.msg = headers
+    def __init__(self, headers: Any) -> None:
+        self.msg: Any = headers
 
-    def isclosed(self):
+    def isclosed(self) -> bool:
         return True
 
-    def close(self):
+    def close(self) -> None:
         return
 
 
 class RequestsMock(object):
-    DELETE = "DELETE"
-    GET = "GET"
-    HEAD = "HEAD"
-    OPTIONS = "OPTIONS"
-    PATCH = "PATCH"
-    POST = "POST"
-    PUT = "PUT"
-    response_callback = None
+    DELETE: Literal["DELETE"] = "DELETE"
+    GET: Literal["GET"] = "GET"
+    HEAD: Literal["HEAD"] = "HEAD"
+    OPTIONS: Literal["OPTIONS"] = "OPTIONS"
+    PATCH: Literal["PATCH"] = "PATCH"
+    POST: Literal["POST"] = "POST"
+    PUT: Literal["PUT"] = "PUT"
+
+    response_callback: Optional[Callable[[Any], Any]] = None
 
     def __init__(
         self,
-        assert_all_requests_are_fired=True,
-        response_callback=None,
-        passthru_prefixes=(),
-        target="requests.adapters.HTTPAdapter.send",
-        registry=FirstMatchRegistry,
+        assert_all_requests_are_fired: bool = True,
+        response_callback: Optional[Callable[[Any], Any]] = None,
+        passthru_prefixes: Tuple[str, ...] = (),
+        target: str = "requests.adapters.HTTPAdapter.send",
+        registry: Type[FirstMatchRegistry] = FirstMatchRegistry,
     ):
-        self._calls = CallList()
+        self._calls: CallList = CallList()
         self.reset()
-        self._registry = registry()  # call only after reset
-        self.assert_all_requests_are_fired = assert_all_requests_are_fired
-        self.response_callback = response_callback
-        self.passthru_prefixes = tuple(passthru_prefixes)
-        self.target = target
-        self._patcher = None
+        self._registry: FirstMatchRegistry = registry()  # call only after reset
+        self.assert_all_requests_are_fired: bool = assert_all_requests_are_fired
+        self.response_callback: Optional[Callable[[Any], Response]] = response_callback
+        self.passthru_prefixes: Tuple[_URLPatternType, ...] = tuple(passthru_prefixes)
+        self.target: str = target
+        self._patcher: Optional["_mock_patcher"] = None
+        self._thread_lock = _ThreadingLock()
 
-    def get_registry(self):
+    def get_registry(self) -> FirstMatchRegistry:
         return self._registry
 
-    def _set_registry(self, new_registry):
+    def _set_registry(self, new_registry: Type[FirstMatchRegistry]) -> None:
         if self.registered():
             err_msg = (
                 "Cannot replace Registry, current registry has responses.\n"
@@ -556,20 +664,20 @@ class RequestsMock(object):
 
         self._registry = new_registry()
 
-    def reset(self):
+    def reset(self) -> None:
         self._registry = FirstMatchRegistry()
         self._calls.reset()
         self.passthru_prefixes = ()
 
     def add(
         self,
-        method=None,  # method or ``Response``
-        url=None,
-        body="",
-        adding_headers=None,
-        *args,
-        **kwargs,
-    ):
+        method: _HTTPMethodOrResponse = None,
+        url: "Optional[_URLPatternType]" = None,
+        body: _Body = "",
+        adding_headers: _HeaderSet = None,
+        *args: Any,
+        **kwargs: Any,
+    ) -> BaseResponse:
         """
         >>> import responses
 
@@ -599,8 +707,7 @@ class RequestsMock(object):
 
         """
         if isinstance(method, BaseResponse):
-            self._registry.add(method)
-            return
+            return self._registry.add(method)
 
         if adding_headers is not None:
             kwargs.setdefault("headers", adding_headers)
@@ -612,30 +719,33 @@ class RequestsMock(object):
                     " Using the `content_type` kwarg is recommended."
                 )
 
-        self._registry.add(Response(method=method, url=url, body=body, **kwargs))
+        assert url is not None
+        assert isinstance(method, str)
+        response = Response(method=method, url=url, body=body, **kwargs)
+        return self._registry.add(response)
 
-    def delete(self, *args, **kwargs):
-        self.add(DELETE, *args, **kwargs)
+    def delete(self, *args: Any, **kwargs: Any) -> BaseResponse:
+        return self.add(DELETE, *args, **kwargs)
 
-    def get(self, *args, **kwargs):
-        self.add(GET, *args, **kwargs)
+    def get(self, *args: Any, **kwargs: Any) -> BaseResponse:
+        return self.add(GET, *args, **kwargs)
 
-    def head(self, *args, **kwargs):
-        self.add(HEAD, *args, **kwargs)
+    def head(self, *args: Any, **kwargs: Any) -> BaseResponse:
+        return self.add(HEAD, *args, **kwargs)
 
-    def options(self, *args, **kwargs):
-        self.add(OPTIONS, *args, **kwargs)
+    def options(self, *args: Any, **kwargs: Any) -> BaseResponse:
+        return self.add(OPTIONS, *args, **kwargs)
 
-    def patch(self, *args, **kwargs):
-        self.add(PATCH, *args, **kwargs)
+    def patch(self, *args: Any, **kwargs: Any) -> BaseResponse:
+        return self.add(PATCH, *args, **kwargs)
 
-    def post(self, *args, **kwargs):
-        self.add(POST, *args, **kwargs)
+    def post(self, *args: Any, **kwargs: Any) -> BaseResponse:
+        return self.add(POST, *args, **kwargs)
 
-    def put(self, *args, **kwargs):
-        self.add(PUT, *args, **kwargs)
+    def put(self, *args: Any, **kwargs: Any) -> BaseResponse:
+        return self.add(PUT, *args, **kwargs)
 
-    def add_passthru(self, prefix):
+    def add_passthru(self, prefix: _URLPatternType) -> None:
         """
         Register a URL prefix or regex to passthru any non-matching mock requests to.
 
@@ -654,7 +764,11 @@ class RequestsMock(object):
             prefix = _clean_unicode(prefix)
         self.passthru_prefixes += (prefix,)
 
-    def remove(self, method_or_response=None, url=None):
+    def remove(
+        self,
+        method_or_response: _HTTPMethodOrResponse = None,
+        url: "Optional[_URLPatternType]" = None,
+    ) -> List[BaseResponse]:
         """
         Removes a response previously added using ``add()``, identified
         either by a response object inheriting ``BaseResponse`` or
@@ -667,11 +781,20 @@ class RequestsMock(object):
         if isinstance(method_or_response, BaseResponse):
             response = method_or_response
         else:
+            assert url is not None
+            assert isinstance(method_or_response, str)
             response = BaseResponse(method=method_or_response, url=url)
 
-        self._registry.remove(response)
+        return self._registry.remove(response)
 
-    def replace(self, method_or_response=None, url=None, body="", *args, **kwargs):
+    def replace(
+        self,
+        method_or_response: _HTTPMethodOrResponse = None,
+        url: "Optional[_URLPatternType]" = None,
+        body: _Body = "",
+        *args: Any,
+        **kwargs: Any,
+    ) -> BaseResponse:
         """
         Replaces a response previously added using ``add()``. The signature
         is identical to ``add()``. The response is identified using ``method``
@@ -684,11 +807,20 @@ class RequestsMock(object):
         if isinstance(method_or_response, BaseResponse):
             response = method_or_response
         else:
+            assert url is not None
+            assert isinstance(method_or_response, str)
             response = Response(method=method_or_response, url=url, body=body, **kwargs)
 
-        self._registry.replace(response)
+        return self._registry.replace(response)
 
-    def upsert(self, method_or_response=None, url=None, body="", *args, **kwargs):
+    def upsert(
+        self,
+        method_or_response: _HTTPMethodOrResponse = None,
+        url: "Optional[_URLPatternType]" = None,
+        body: _Body = "",
+        *args: Any,
+        **kwargs: Any,
+    ) -> BaseResponse:
         """
         Replaces a response previously added using ``add()``, or adds the response
         if no response exists.  Responses are matched using ``method``and ``url``.
@@ -699,19 +831,21 @@ class RequestsMock(object):
         >>> responses.upsert(responses.GET, 'http://example.org', json={'data': 2})
         """
         try:
-            self.replace(method_or_response, url, body, *args, **kwargs)
+            return self.replace(method_or_response, url, body, *args, **kwargs)
         except ValueError:
-            self.add(method_or_response, url, body, *args, **kwargs)
+            return self.add(method_or_response, url, body, *args, **kwargs)
 
     def add_callback(
         self,
-        method,
-        url,
-        callback,
-        match_querystring=FalseBool(),
-        content_type="text/plain",
-        match=(),
-    ):
+        method: str,
+        url: _URLPatternType,
+        callback: Callable[
+            ["PreparedRequest"], Union[Exception, Tuple[int, Mapping[str, str], _Body]]
+        ],
+        match_querystring: Union[bool, FalseBool] = FalseBool(),
+        content_type: Optional[str] = "text/plain",
+        match: "_MatcherIterable" = (),
+    ) -> None:
 
         self._registry.add(
             CallbackResponse(
@@ -724,33 +858,66 @@ class RequestsMock(object):
             )
         )
 
-    def registered(self):
+    def registered(self) -> List["BaseResponse"]:
         return self._registry.registered
 
     @property
-    def calls(self):
+    def calls(self) -> CallList:
         return self._calls
 
-    def __enter__(self):
+    def __enter__(self) -> "RequestsMock":
         self.start()
         return self
 
-    def __exit__(self, type, value, traceback):
+    def __exit__(self, type: Any, value: Any, traceback: Any) -> bool:
         success = type is None
         self.stop(allow_assert=success)
         self.reset()
         return success
 
-    def activate(self, func=None, registry=None):
+    @overload
+    def activate(self, func: _F = ...) -> _F:
+        """Overload for scenario when 'responses.activate' is used."""
+        ...  # pragma: no cover
+
+    @overload
+    def activate(
+        self,
+        *,
+        registry: Type[Any] = ...,
+        assert_all_requests_are_fired: bool = ...,
+    ) -> Callable[["_F"], "_F"]:
+        """Overload for scenario when
+        'responses.activate(registry=, assert_all_requests_are_fired=True)' is used.
+
+        See https://github.com/getsentry/responses/pull/469 for more details
+
+        """
+        ...  # pragma: no cover
+
+    def activate(
+        self,
+        func: Optional[_F] = None,
+        *,
+        registry: Optional[Type[Any]] = None,
+        assert_all_requests_are_fired: bool = False,
+    ) -> Union[Callable[["_F"], "_F"], _F]:
         if func is not None:
             return get_wrapped(func, self)
 
-        def deco_activate(function):
-            return get_wrapped(function, self, registry)
+        def deco_activate(function: _F) -> Callable[..., Any]:
+            return get_wrapped(
+                function,
+                self,
+                registry=registry,
+                assert_all_requests_are_fired=assert_all_requests_are_fired,
+            )
 
         return deco_activate
 
-    def _find_match(self, request):
+    def _find_match(
+        self, request: "PreparedRequest"
+    ) -> Tuple[Optional["BaseResponse"], List[str]]:
         """
         Iterates through all available matches and validates if any of them matches the request
 
@@ -759,22 +926,33 @@ class RequestsMock(object):
             (Response) found match. If multiple found, then remove & return the first match.
             (list) list with reasons why other matches don't match
         """
-        return self._registry.find(request)
+        with self._thread_lock:
+            return self._registry.find(request)
 
-    def _parse_request_params(self, url):
-        params = {}
+    def _parse_request_params(
+        self, url: str
+    ) -> Dict[str, Union[str, int, float, List[Optional[Union[str, int, float]]]]]:
+        params: Dict[str, Union[str, int, float, List[Any]]] = {}
         for key, val in groupby(parse_qsl(urlsplit(url).query), lambda kv: kv[0]):
             values = list(map(lambda x: x[1], val))
             if len(values) == 1:
-                values = values[0]
+                values = values[0]  # type: ignore[assignment]
             params[key] = values
         return params
 
-    def _on_request(self, adapter, request, **kwargs):
+    def _on_request(
+        self,
+        adapter: "HTTPAdapter",
+        request: "PreparedRequest",
+        *,
+        retries: Optional["_Retry"] = None,
+        **kwargs: Any,
+    ) -> "models.Response":
         # add attributes params and req_kwargs to 'request' object for further match comparison
         # original request object does not have these attributes
-        request.params = self._parse_request_params(request.path_url)
-        request.req_kwargs = kwargs
+        request.params = self._parse_request_params(request.path_url)  # type: ignore[attr-defined]
+        request.req_kwargs = kwargs  # type: ignore[attr-defined]
+        request_url = str(request.url)
 
         match, match_failed_reasons = self._find_match(request)
         resp_callback = self.response_callback
@@ -782,20 +960,20 @@ class RequestsMock(object):
         if match is None:
             if any(
                 [
-                    p.match(request.url)
+                    p.match(request_url)
                     if isinstance(p, Pattern)
-                    else request.url.startswith(p)
+                    else request_url.startswith(p)
                     for p in self.passthru_prefixes
                 ]
             ):
-                logger.info("request.allowed-passthru", extra={"url": request.url})
+                logger.info("request.allowed-passthru", extra={"url": request_url})
                 return _real_send(adapter, request, **kwargs)
 
             error_msg = (
                 "Connection refused by Responses - the call doesn't "
                 "match any registered mock.\n\n"
                 "Request: \n"
-                f"- {request.method} {request.url}\n\n"
+                f"- {request.method} {request_url}\n\n"
                 "Available matches:\n"
             )
             for i, m in enumerate(self.registered()):
@@ -815,35 +993,58 @@ class RequestsMock(object):
             raise response
 
         if match.passthrough:
-            logger.info("request.passthrough-response", extra={"url": request.url})
-            response = _real_send(adapter, request, **kwargs)
+            logger.info("request.passthrough-response", extra={"url": request_url})
+            response = _real_send(adapter, request, **kwargs)  # type: ignore[assignment]
         else:
             try:
-                response = adapter.build_response(request, match.get_response(request))
+                response = adapter.build_response(  # type: ignore[no-untyped-call]
+                    request, match.get_response(request)
+                )
             except BaseException as response:
                 match.call_count += 1
                 self._calls.add(request, response)
                 raise
 
-        response = resp_callback(response) if resp_callback else response
+        if resp_callback:
+            response = resp_callback(response)  # type: ignore[misc]
         match.call_count += 1
-        self._calls.add(request, response)
+        self._calls.add(request, response)  # type: ignore[misc]
+
+        retries = retries or adapter.max_retries
+        # first validate that current request is eligible to be retried.
+        # See ``requests.packages.urllib3.util.retry.Retry`` documentation.
+        if retries.is_retry(
+            method=response.request.method, status_code=response.status_code  # type: ignore[misc]
+        ):
+            try:
+                retries = retries.increment(
+                    method=response.request.method,  # type: ignore[misc]
+                    url=response.url,  # type: ignore[misc]
+                    response=response,  # type: ignore[misc]
+                )
+                return self._on_request(adapter, request, retries=retries, **kwargs)
+            except MaxRetryError:
+                if retries.raise_on_status:
+                    raise
+                return response
         return response
 
-    def start(self):
+    def start(self) -> None:
         if self._patcher:
             # we must not override value of the _patcher if already applied
             # this prevents issues when one decorated function is called from
             # another decorated function
             return
 
-        def unbound_on_send(adapter, request, *a, **kwargs):
+        def unbound_on_send(
+            adapter: "HTTPAdapter", request: "PreparedRequest", *a: Any, **kwargs: Any
+        ) -> "models.Response":
             return self._on_request(adapter, request, *a, **kwargs)
 
         self._patcher = std_mock.patch(target=self.target, new=unbound_on_send)
         self._patcher.start()
 
-    def stop(self, allow_assert=True):
+    def stop(self, allow_assert: bool = True) -> None:
         if self._patcher:
             # prevent stopping unstarted patchers
             self._patcher.stop()
@@ -866,7 +1067,7 @@ class RequestsMock(object):
                 )
             )
 
-    def assert_call_count(self, url, count):
+    def assert_call_count(self, url: str, count: int) -> bool:
         call_count = len(
             [
                 1
@@ -959,7 +1160,7 @@ upsert = _default_mock.upsert
 deprecated_names = ["assert_all_requests_are_fired", "passthru_prefixes", "target"]
 
 
-def __getattr__(name):
+def __getattr__(name: str) -> Any:
     if name in deprecated_names:
         warn(
             f"{name} is deprecated. Please use 'responses.mock.{name}",
